#!/usr/bin/python3

import rclpy
import rclpy.qos
from threading import Thread

import frc_robot_utilities_py_node.frc_robot_utilities_py as frc_utils
from frc_robot_utilities_py_node.RobotStatusHelperPy import RobotStatusHelperPy, Alliance, RobotMode
from frc_robot_utilities_py_node.BufferedROSMsgHandlerPy import BufferedROSMsgHandlerPy
from joystick_simulation_node.generated.parameters import ParameterizedNode

import sys, signal
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from ck_ros2_msgs_node.msg import *
from ck_ros2_base_msgs_node.msg import *
from ck_utilities_py_node.node_handle import NodeHandle

from ck_utilities_py_node.pyqt5_signal_wakeup_handler import SignalWakeupHandler

from joystick_simulation_node.controllers.button_box_controller import ButtonBoxController
from joystick_simulation_node.controllers.ruffy_controller import RuffyController
from joystick_simulation_node.controllers.extreme3d_controller import Extreme3DController
# from joystick_simulation_node.controllers.xbox_controller import XBoxController
from joystick_simulation_node.controllers.robot_status_controller import RobotStatusController, State, Alliance
from joystick_simulation_node.controllers.components.generic_controller import GenericController

from typing import List

joysticks : List[GenericController] = []
robot_status_controller : RobotStatusController = None

class MainWindow(QMainWindow):
    def __init__(self, show_operator_controls: bool):
        global joysticks
        global robot_status_controller

        super().__init__()

        self.setWindowTitle('Joystick Simulator')
        icon = self.style().standardIcon(QStyle.SP_TitleBarMenuButton)
        self.setWindowIcon(icon)

        cw = QWidget(self)
        self.setCentralWidget(cw)
        ml = QHBoxLayout(cw)
        cw.setLayout(ml)

        robot_status_controller = RobotStatusController()
        robot_status_controller.setParent(cw)
        ml.addWidget(robot_status_controller)

        driver_joystick = RuffyController("driver_joystick", 0)
        driver_joystick.setParent(cw)
        ml.addWidget(driver_joystick)
        joysticks.append(driver_joystick)

        # TODO (Chris): Read bools from params to enable/disable operator controls
        if show_operator_controls:
            buttonbox = ButtonBoxController("button_box_1", 2, 12)
            buttonbox.setParent(cw)
            ml.addWidget(buttonbox)
            joysticks.append(buttonbox)

            arm_joystick = Extreme3DController("arm_joystick", 1)
            arm_joystick.setParent(cw)
            ml.addWidget(arm_joystick)
            joysticks.append(arm_joystick)

class LocalNode(ParameterizedNode):
    def __init__(self):
        super().__init__('joystick_simulation_node')

        self.joystickStatusPublisher = self.create_publisher(JoystickStatusArray, '/JoystickStatus', rclpy.qos.QoSProfile(depth=50, durability=rclpy.qos.QoSDurabilityPolicy.VOLATILE))
        self.robotStatusPublisher = self.create_publisher(RobotStatus, '/RobotStatus', rclpy.qos.QoSProfile(depth=50, durability=rclpy.qos.QoSDurabilityPolicy.VOLATILE))
        self.autoSelectionPublisher = self.create_publisher(AutonomousSelection, '/AutonomousSelection', rclpy.qos.QoSProfile(depth=50, durability=rclpy.qos.QoSDurabilityPolicy.VOLATILE))
        self.autoConfigSubscriber = BufferedROSMsgHandlerPy(AutonomousConfiguration)
        self.autoConfigSubscriber.register_for_updates("/AutonomousConfiguration")

        self.auto_options = []

        self.create_timer(1.0 / 100.0, self.ros_func)

    def ros_func(self):
        global joysticks
        global robot_status_controller

        if not rclpy.ok():
            return

        joystick_status_array = JoystickStatusArray()

        for controller in joysticks:
            if controller is not None:
                joystick_msg = JoystickStatus()
                joystick_msg.index = controller.index
                joystick_msg.axes = controller.controller_data.axes
                joystick_msg.buttons = controller.controller_data.buttons
                joystick_msg.povs = controller.controller_data.povs
                joystick_status_array.joysticks.append(joystick_msg)

        self.joystickStatusPublisher.publish(joystick_status_array)

        override_robot_status_msg = RobotStatus()

        if robot_status_controller is not None:
            if robot_status_controller.robot_enabled:
                if robot_status_controller.robot_state == State.AUTO:
                    override_robot_status_msg.robot_state = RobotStatus.AUTONOMOUS
                elif robot_status_controller.robot_state == State.TELEOP:
                    override_robot_status_msg.robot_state = RobotStatus.TELEOP
                elif robot_status_controller.robot_state == State.TEST:
                    override_robot_status_msg.robot_state = RobotStatus.TEST
                else:
                    override_robot_status_msg.robot_state = RobotStatus.DISABLED
            else:
                override_robot_status_msg.robot_state = RobotStatus.DISABLED

            override_robot_status_msg.alliance = RobotStatus.RED if robot_status_controller.robot_alliance == Alliance.RED else RobotStatus.BLUE
            override_robot_status_msg.event_name = ''
            override_robot_status_msg.game_data = ''
            override_robot_status_msg.is_connected = True
            override_robot_status_msg.match_number = 0
            override_robot_status_msg.match_time = -1.0
            override_robot_status_msg.match_type = RobotStatus.PRACTICE
            override_robot_status_msg.station = 1
            override_robot_status_msg.system_time = ''

            auto_config: AutonomousConfiguration = self.autoConfigSubscriber.get()

            if self.Parameters.enable_auto_selection:
                if auto_config is not None and auto_config.autonomous_options != self.auto_options:
                    self.auto_options = auto_config.autonomous_options
                    robot_status_controller.set_auto_options(self.auto_options)

                selected_auto = robot_status_controller.get_selected_auto()

                if selected_auto != '':
                    auto_selection = AutonomousSelection()
                    auto_selection.autonomous = selected_auto
                    self.autoSelectionPublisher.publish(auto_selection)


        self.robotStatusPublisher.publish(override_robot_status_msg)

ros_thread = None
def quit_handler(signal, frame):
    rclpy.shutdown()
    ros_thread.join(5)
    QApplication.quit()
    sys.exit(0)

def window(show_operator_controls: bool):
    app = QApplication(sys.argv)
    widget = MainWindow(show_operator_controls)
    widget.show()

    SignalWakeupHandler(app)

    signal.signal(signal.SIGINT, quit_handler)

    sys.exit(app.exec())

def main(args=None):
    global ros_thread
    rclpy.init(args=args)
    node = LocalNode()

    def spin_target():
        rclpy.spin(node)

    ros_thread = Thread(target=spin_target)
    ros_thread.start()

    window(node.Parameters.show_operator_controls)


if __name__ == '__main__':
    main()